
__version__ = "0.12rc4"
__version_vcs_info__ = {
    'branch': 'release-0.12',
    'revid': '16a6ffc',
    'revno': 1890}
__version_user_email__ = "saxtouri@admin.grnet.gr"
__version_user_name__ = "Stavros Sachtouris"
